/******************************************************************************
 * sample.c
 * SASS Compiler for CoreTone, sample descriptor parsing and pack generation.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in ctcc.c
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include "../types.h"
#include "../CoreTone/sample.h"
#include "../CoreTone/channel.h"
#include "../CoreTone/music.h"

#include "useful.h"
#include "sample.h"

/******************************************************************************
 * Text Tables
 ******************************************************************************/
extern const char szERR_NoMainMem[];
extern const char szERR_EarlyEnd[];
extern const char szERR_NoOpen[];
extern const char szERR_NoDest[];

extern const char szERR_Garbage[];

const char szSMP_Parse[] =		"Parsing sample descriptor file : ";
const char szSMP_Make[] =		"Saving sample package to : ";

const char szSMP_Empty[] =		"WARNING- No samples defined, are you sure about this?\n";

const char szSMP_TwoLabels[] =	"ERROR- Sample already defined with label : ";

const char szSMP_NoFreqS[] =	"ERROR- No sample frequency for definition of : ";
const char szSMP_BadFreqS[] =	"ERROR- Unable to parse sample frequency : ";
const char szSMP_RangeFreqS[] =	"ERROR- Sample frequency out of range (1.0 - 65535.99) : ";
const char szSMP_NoFreqC[] =	"ERROR- No content frequency for definition of : ";
const char szSMP_BadFreqC[] =	"ERROR- Unable to parse content frequency : ";
const char szSMP_RangeFreqC[] =	"ERROR- Content frequency out of range (1.0 - 65535.99) : ";

const char szSMP_BadOffS[] =	"ERROR- Unable to parse start offset : ";
const char szSMP_RangeOffS[] =	"ERROR- Start offset out of range (0 - 65535) : ";
const char szSMP_BadOffE[] =	"ERROR- Unable to parse end offset : ";
const char szSMP_RangeOffE[] =	"ERROR- End offset out of range (0 - 65535) : ";
const char szSMP_SwapOff[] =	"ERROR- Sample start and end offsets reversed in : ";
const char szSMP_ZeroSample[] =	"ERROR- Sample start and end offsets equal in : ";
const char szSMP_GenRange[] =	"ERROR- Sample start and end offsets out of range for : ";

const char szSMP_BadWrite[] =	"ERROR- Couldn't write to sample pack file : ";

const char szSMP_WhiteSpace[] =	":{}()[], \t\n";

/******************************************************************************
 * Sample Descriptor Table
 ******************************************************************************/
Sample_t *pSamples = NULL;
uint32_t uiSamples = 0, uiSampleSlots = 0;


/******************************************************************************
 * !!!!----             SASS SAMPLE DESCRIPTOR PARSING               ----!!!!
 ******************************************************************************/
/* int32_t smp_parse(FILE *pDesc, char *pszFileName)
 *  Parse the sample descriptor file pDesc, using its contents to populate the
 * sample descriptor table which may be used to verify instrument and sound
 * effect definitions or generate a sample pack binary. The source file pDesc
 * should already be open when this function is called.
 *  Returns a nonzero value if anything went wrong during parsing.
 *----------------------------------------------------------------------------*/
int32_t smp_parse(FILE *pDesc, char *pszFileName)
{
	char *pszLineBuf = NULL;
	char *pszMainTok = NULL,*pszSubTok = NULL;
	uint32_t uiCurLine;

	int32_t iAborted = -1,iStatus = 0;
	uint32_t uiX,uiY,uiZ;


	/**
	 * !!!!---- PARSE START ----!!!!
	 */
	fprintf(stdout, "%s%s\n", szSMP_Parse, pszFileName);
	pSamples = malloc(sizeof(Sample_t) * SAMPLE_DTABLE_BASE_LEN);
	uiSamples = 0;
	uiSampleSlots = SAMPLE_DTABLE_BASE_LEN;

	pszLineBuf = malloc(sizeof(char) * SAMPLE_PARSER_LINEBUF_LEN);
	if((NULL == pSamples) || (NULL == pszLineBuf))
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		goto cleanup;
	}


	/**
	 * Samples are expected to be defined one per line as follows,
	 *
	 * Name : (Sample File) (Sample Frequency) (Content Frequency) ...
	 *        [Start] [End]
	 *
	 * Where the last two arguments (start and end) are optional, allowing
	 * the user to reuse portions of one large sample in several smaller ones.
	 */
	for(uiCurLine = 1; ; uiCurLine++)
	{
		if(fgets(pszLineBuf, SAMPLE_PARSER_LINEBUF_LEN, pDesc) == NULL)
		{
			iAborted = 0;
			break;
		}

		if(strcut(pszLineBuf, ";", '\0'))
			break;

		pszMainTok = strtok(pszLineBuf, szSMP_WhiteSpace);
		pszSubTok = strtok(NULL, szSMP_WhiteSpace);
		if((NULL != pszMainTok) && (NULL != pszSubTok))
		{
			/**
			 * !!!!---- NAME AND SOURCE FILE ----!!!!
			 * Ensure the sample's name hasn't already been used in the
			 * descriptor table, and that there is enough room to contain
			 * its new entry.
			 */
			if(0 == smp_get(pszMainTok, NULL, NULL))
			{
				fprintf(stderr, "%s%s in : %s,%u\n", szSMP_TwoLabels,
					pszMainTok, pszFileName, uiCurLine);
				goto cleanup;
			}

			if(tealloc(&pSamples, &uiSampleSlots, (uiSamples + 1),
				sizeof(Sample_t))) goto cleanup;

			/**
			 * Add the new sample descriptor entry, copying in the sample's
			 * name and source file (not verified to be an actual file until
			 * we attempt to build a sample package).
			 */
			uiX = uiSamples;
			uiY = strlen(pszMainTok) + 1;
			uiZ = strlen(pszSubTok) + 1;
			pSamples[uiX].pszName = malloc(sizeof(char) * uiY);
			pSamples[uiX].pszFile = malloc(sizeof(char) * uiZ);
			pSamples[uiX].uiStart = 0;
			pSamples[uiX].uiEnd = SAMPLE_PARSER_END_TAG;
			pSamples[uiX].uiOffset = 0;
			pSamples[uiX].uiLength = 0;
			pSamples[uiX].pData = NULL;
			pSamples[uiX].iSkipLoad = 0;
			uiSamples++;

			if((NULL == pSamples[uiX].pszName) || (NULL == pSamples[uiX].pszFile))
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				goto cleanup;
			}

			strcpy(pSamples[uiX].pszName, pszMainTok);
			strcpy(pSamples[uiX].pszFile, pszSubTok);


			/**
			 * !!!!---- SAMPLE AND CONTENT FREQUENCIES ----!!!!
			 * Get the sample and content frequency, both of these may
			 * range from 1.0Hz to 65535.99Hz (stored as 16.16 precision
			 * in the actual sample packs).
			 */
			if(NULL == (pszSubTok = strtok(NULL, szSMP_WhiteSpace)))
			{
				fprintf(stderr, "%s%s in : %s,%u\n", szSMP_NoFreqS,
					pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if(1 != sscanf(pszSubTok, "%lf", &(pSamples[uiX].dSf)))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_BadFreqS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if((SAMPLE_PARSER_FREQ_MIN > pSamples[uiX].dSf)
				|| (SAMPLE_PARSER_FREQ_MAX < pSamples[uiX].dSf))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_RangeFreqS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}

			
			if(NULL == (pszSubTok = strtok(NULL, szSMP_WhiteSpace)))
			{
				fprintf(stderr, "%s%s in : %s,%u\n", szSMP_NoFreqC,
					pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if(1 != sscanf(pszSubTok, "%lf", &(pSamples[uiX].dBf)))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_BadFreqC,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if((SAMPLE_PARSER_FREQ_MIN > pSamples[uiX].dBf)
				|| (SAMPLE_PARSER_FREQ_MAX < pSamples[uiX].dBf))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_RangeFreqC,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}

			
			/**
			 * !!!!---- START AND END OFFSETS ----!!!!
			 * Both of these parameters are optional, allowing the user to
			 * specify a region within a file to use as the sample's data.
			 *
			 * Specifying a start index without an end is valid and will
			 * just clip the beginning of the sample.
			 */
			if(NULL == (pszSubTok = strtok(NULL, szSMP_WhiteSpace)))
				continue;
			if(1 != sscanf(pszSubTok, "%d", &(pSamples[uiX].uiStart)))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_BadOffS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if((SAMPLE_PARSER_LENGTH_MIN > pSamples[uiX].uiStart)
				|| (SAMPLE_PARSER_LENGTH_MAX <= pSamples[uiX].uiStart))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_RangeOffS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}


			if(NULL == (pszSubTok = strtok(NULL, szSMP_WhiteSpace)))
				continue;
			if(1 != sscanf(pszSubTok, "%d", &(pSamples[uiX].uiEnd)))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_BadOffS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}
			if((SAMPLE_PARSER_LENGTH_MIN > pSamples[uiX].uiEnd)
				|| (SAMPLE_PARSER_LENGTH_MAX <= pSamples[uiX].uiEnd))
			{
				fprintf(stderr, "%s%s of : %s in : %s,%u\n", szSMP_RangeOffS,
					pszSubTok, pSamples[uiX].pszName, pszFileName, uiCurLine);
				goto cleanup;
			}

			if(pSamples[uiX].uiEnd < pSamples[uiX].uiStart)
			{
				fprintf(stderr, "%s%s,%u\n", szSMP_SwapOff,
					pszFileName, uiCurLine);
				goto cleanup;
			}
			else if(pSamples[uiX].uiStart == pSamples[uiX].uiEnd)
			{
				fprintf(stderr, "%s%s,%u\n", szSMP_ZeroSample,
					pszFileName, uiCurLine);
				goto cleanup;
			}

			/**
			 * If we've made it this far, there could be garbage at the
			 * end of our current line. Let's check and see.
			 */
			pszSubTok = strtok(NULL, szSMP_WhiteSpace);
			if(NULL != pszSubTok)
			{
				fprintf(stderr, "%s%s,%u\n",
					szERR_Garbage, pszFileName, uiCurLine);
				break;
			}
		}
	}

	if(0 == uiSamples)
	{
		fprintf(stderr, "%s", szSMP_Empty);
	}

	/**
	 * !!!!---- PARSE TEARDOWN ----!!!!
	 * Release everything EXCLUDING the sample descriptor table (pSamples),
	 * this can be freed later using smp_close().
	 */
cleanup:
	if(NULL != pszLineBuf) free(pszLineBuf);

	return (iStatus | iAborted);
}


/* int32_t smp_make(FILE *pDest, char *pszFileName)
 *  Attempt to build and write a sample package for the currently populated
 * sample descriptor table. The destination file pDest should be open for
 * binary writing at the time this function is called.
 *  Returns a nonzero value if anything went wrong during the build.
 *----------------------------------------------------------------------------*/
int32_t smp_make(FILE *pDest, char *pszFileName)
{
	FILE *pCurFile = NULL;
	uint8_t *pPack = NULL;
	uint32_t *pDir = NULL;

	uint32_t uiFiles,uiLen;
	uint32_t uiFileBase;

	int16p16_t iSf,iBf;
	double dFr;
	double dFr_int,dFr_frac;

	uint32_t uiX,uiY,uiZ;
	int32_t iStatus = -1;

	/**
	 * !!!!---- INCLUSION PRESCAN ----!!!!
	 * As the user is allowed to define multiple samples from one included file,
	 * we need to detect the first instance of said file in our sample
	 * descriptor table (counting upward from entry zero).
	 *
	 * While files will always be included in full, we only want to include
	 * them once. Each sample's iSkipLoad will be clear if and only if said
	 * sample contains the first mention of a file, and set if the file has
	 * already been mentioned by a previous sample in the list.
	 * In the latter case uiFileNum is set to the index of said sample
	 * descriptor, and the sample will use the offset and length (uiOffset and
	 * uiLength) of this descriptor to calculate its own position in the pack.
	 *
	 * Oh right, we're also accumulating the header + directory length for
	 * the sample pack into uiLen during all this.
	 */
	fprintf(stdout, "%s%s\n", szSMP_Make, pszFileName);

	uiFiles = 0;
	uiLen = CORETONE_SMPPAK_HEAD_SIZE;
	for(uiX = 0; uiX < uiSamples; uiX++)
	{
		for(uiY = 0; uiY < uiX; uiY++)
		{
			if(0 == strcmp(pSamples[uiX].pszFile, pSamples[uiY].pszFile))
			{
				pSamples[uiX].uiFileNum = uiY;
				pSamples[uiX].iSkipLoad = -1;
				break;
			}
		}

		if(uiY == uiX)
		{
			pSamples[uiX].uiFileNum = uiX;
			uiFiles++;
		}

		uiLen += (CORETONE_SMPPAK_ENTRY_SIZE * sizeof(uint32_t));
	}

	uiFileBase = uiLen;

	/**
	 * !!!!---- SOURCE DATA FETCH ----!!!!
	 * Now that we know the sample pack directory length and the true number
	 * of included files, we can ensure each file exists, get its size
	 * to calculate the final package size and sample offsets, and load in
	 * its data.
	 */
	for(uiX = 0; uiX < uiSamples; uiX++)
	{
		if(pSamples[uiX].iSkipLoad)
		{
			/**
			 * If the sample has been tagged as a secondary usage of an
			 * included file, pull its offset and length from the first
			 * inclusion (uiFileNum).
			 */
			uiY = pSamples[uiX].uiFileNum;
			pSamples[uiX].uiOffset = pSamples[uiY].uiOffset;
			pSamples[uiX].uiLength = pSamples[uiY].uiLength;
		}
		else
		{
			/**
			 * Otherwise, this descriptor is the first usage of its
			 * specified file. Let's try and open it, calculate its
			 * size and offset, then load it up.
			 */
			pSamples[uiX].uiOffset = uiLen;
			if(NULL == (pCurFile = fopen(pSamples[uiX].pszFile, "rb")))
			{
				fprintf(stderr, "%s%s\n", szERR_NoOpen, pSamples[uiX].pszFile);
				goto cleanup;
			}

			fseek(pCurFile, 0, SEEK_END);
			pSamples[uiX].uiLength = ftell(pCurFile);
			rewind(pCurFile);
			if(NULL == (pSamples[uiX].pData = malloc(sizeof(uint8_t) * pSamples[uiX].uiLength)))
			{
				fprintf(stderr, "%s", szERR_NoMainMem);
				goto cleanup;
			}

			uiZ = fread(pSamples[uiX].pData, sizeof(uint8_t), pSamples[uiX].uiLength, pCurFile);
			if(uiZ != pSamples[uiX].uiLength)
			{
				fprintf(stderr, "%s%s\n", szERR_EarlyEnd, pSamples[uiX].pszFile);
				goto cleanup;
			}

			/**
			 * Alright, the file has been loaded in. As a last step we need
			 * to accumulate its length into the final pack size (uiLen).
			 */
			fclose(pCurFile);
			pCurFile = NULL;
			uiLen += pSamples[uiX].uiLength;
		}

		/**
		 * Verify the sample's start and end offsets (if manually specified by
		 * the user). Both must fall within the source data and default to the
		 * first (start, inclusive) and last (end, exclusive) samples.
		 */
		if(SAMPLE_PARSER_END_TAG == pSamples[uiX].uiEnd)
		{
			pSamples[uiX].uiEnd = pSamples[uiX].uiLength;
		}

		if((pSamples[uiX].uiStart >= pSamples[uiX].uiLength)
			|| (pSamples[uiX].uiEnd > pSamples[uiX].uiLength))
		{
			fprintf(stderr, "%s%s\n", szSMP_GenRange, pSamples[uiX].pszFile);
			goto cleanup;
		}
	}

	/**
	 * !!!!---- SAMPLE PACK ASSEMBLE + WRITE ----!!!!
	 * Our sample descriptors have all been verified and any pertinent source
	 * files loaded up, we can now build the final sample pack.
	 *
	 * I'm making the assumption we'll get a 32-Bit aligned block of memory
	 * from malloc, which should be fair.
	 */
	if(NULL == (pPack = malloc(sizeof(uint8_t) * uiLen)))
	{
		fprintf(stderr, "%s", szERR_NoMainMem);
		goto cleanup;
	}

	strcpy(pPack, CORETONE_SMPPAK_HEAD_MAGICWORD);
	memcpy(pPack + CORETONE_SMPPAK_HEAD_COUNT, &uiSamples, sizeof(uint32_t));
	pDir = (uint32_t*)(pPack + CORETONE_SMPPAK_HEAD_SIZE);
	for(uiX = 0; uiX < uiSamples; uiX++)
	{
		if(!(pSamples[uiX].iSkipLoad))
		{
			memcpy((pPack + pSamples[uiX].uiOffset),
				pSamples[uiX].pData, (sizeof(uint8_t) * pSamples[uiX].uiLength));
		}

		dFr = pSamples[uiX].dSf;
		dFr_int = floor(dFr);
		dFr_frac = dFr - dFr_int;
		dFr_frac *= 65536.0;
		iSf.sPair.sHi = (int16_t)dFr_int;
		iSf.usPair.usLo = (uint16_t)dFr_frac;

		dFr = pSamples[uiX].dBf;
		dFr_int = floor(dFr);
		dFr_frac = dFr - dFr_int;
		dFr_frac *= 65536.0;
		iBf.sPair.sHi = (int16_t)dFr_int;
		iBf.usPair.usLo = (uint16_t)dFr_frac;

		pDir[CORETONE_SMPPAK_ENTRY_OFF] =
			pSamples[uiX].uiOffset + pSamples[uiX].uiStart;
		pDir[CORETONE_SMPPAK_ENTRY_SLEN] =
			pSamples[uiX].uiEnd - pSamples[uiX].uiStart;
		pDir[CORETONE_SMPPAK_ENTRY_SFREQ] = iSf.uiWhole;
		pDir[CORETONE_SMPPAK_ENTRY_BFREQ] = iBf.uiWhole;

		pDir += CORETONE_SMPPAK_ENTRY_SIZE;
	}

	uiZ = fwrite(pPack, sizeof(uint8_t), uiLen, pDest);
	if(uiLen != uiZ)
	{
		fprintf(stderr, "%s%s\n", szSMP_BadWrite, pszFileName);
		goto cleanup;
	}

	fprintf(stdout, "\nSuccess! %u Bytes written to : %s\n", uiZ, pszFileName);
	fprintf(stdout, "Total Samples : %u, Included Files : %u\n",
		uiSamples, uiFiles);
	iStatus = 0;
	fflush(pDest);

cleanup:
	if(NULL != pCurFile) fclose(pCurFile);
	if(NULL != pPack) free(pPack);

	return iStatus;
}


/* void smp_close(void)
 *  Close the sample descriptor table, call before the program exits.
 *----------------------------------------------------------------------------*/
void smp_close(void)
{
	uint32_t uiX;

	if(NULL != pSamples)
	{
		for(uiX = 0; uiX < uiSamples; uiX++)
		{
			if(NULL != pSamples[uiX].pszName)
			{
				free(pSamples[uiX].pszName);
				pSamples[uiX].pszName = NULL;
			}
			if(NULL != pSamples[uiX].pszFile)
			{
				free(pSamples[uiX].pszFile);
				pSamples[uiX].pszFile = NULL;
			}

			if(NULL != pSamples[uiX].pData)
			{
				free(pSamples[uiX].pData);
				pSamples[uiX].pData = NULL;
			}
		}

		free(pSamples);
		pSamples = NULL;
	}
}




/******************************************************************************
 * !!!!----                  SAMPLE TABLE ACCESS                     ----!!!!
 ******************************************************************************/
/* int32_t smp_get(char *pszName, uint32_t *puiNum, Sample_t **ppSample)
 *  Attempt to locate and fetch a sample with the name pszName, returning
 * its index and descriptor entry in *puiNum and *ppSample, respectively.
 *
 *  Returns a nonzero value if the sample could not be located.
 *----------------------------------------------------------------------------*/
int32_t smp_get(char *pszName, uint32_t *puiNum, Sample_t **ppSample)
{
	int32_t iFailed = -1;
	uint32_t uiX;

	if(NULL != pSamples)
	{
		for(uiX = 0; uiX < uiSamples; uiX++)
		{
			if(0 == strcmp(pszName, pSamples[uiX].pszName))
			{
				if(NULL != puiNum)
					*puiNum = uiX;
				if(NULL != ppSample)
					*ppSample = pSamples + uiX;

				iFailed = 0;
				break;
			}
		}
	}
	return iFailed;
}
